"""Utilities for working with the file system."""
import os
import time
import pathlib

import inductiva
from inductiva import types


def find_path_to_package(package_dir: str):
    """Find path to package directory in inductiva library in local computer.
    
    Args:
        package_dir: Name of package directory.
    """
    return os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", package_dir))


def get_timestamped_path(path: types.Path, sep: str = "-") -> pathlib.Path:
    """Return a path that does not exist by appending a timestamp.

    Args:
        path: Path to a file or directory.

    Returns:
        A path that does not exist by appending the timestamp.
    """
    path = pathlib.Path(path)
    timestamp = time.strftime("%Y-%m-%dT%Hh%Mm%Ss")

    name = f"{path.stem}{sep}{timestamp}"

    return path.with_name(name + path.suffix)


def resolve_path(path: types.Path) -> pathlib.Path:
    """Resolve a path relative to the Inductiva package working directory.

    Args:
        path: Path to a file or directory.
    """
    root = pathlib.Path.cwd()

    if inductiva.working_dir:
        root = pathlib.Path(inductiva.working_dir)

    return pathlib.Path(root, path)


def get_sorted_files(data_dir: str,
                     file_format: str = "name",
                     split_token: str = "_"):
    """Returns list of files sorted according to [file_key].

    Order a set of .format files of the form
    ['name_1.format', 'name_2.format',...,'name_10.format',
    ...,'name_n.format'].

    The default sorting methods for list, list.sort()
    or sorted(list), order 'name_10.format' before 'name_2.format',
    which is not representative of the time series.

    In this function we sort according to the number..
    """

    if not os.path.exists(data_dir):
        raise IOError(f"Directory '{data_dir}' does not exist.")

    # Get a list of the files in the data directory.
    files = os.scandir(data_dir)

    # The files have file_format extension.
    files = [
        file for file in files if pathlib.Path(file.path).suffix == file_format
    ]

    # Sort the files to be read according to [file_key].
    def get_alphanum_key(file):
        file_name = pathlib.Path(file.path).stem
        file_name_splits = file_name.split(split_token)
        file_key = file_name_splits[-1]
        return int(file_key)

    files = sorted(files, key=get_alphanum_key)

    return files


def remove_files_with_tag(main_dir: str, remove_tag: str) -> None:
    """Remove files in a directory that contain a specific name tag.
    
    Iterates over all folders in the main folder and remove all files
    that contain remove_tag as a substring in the filename.

    Args:
        main_dir: Path to a directory.
        remove_tag: tag that identifies all files to be removed."""

    for item in os.listdir(main_dir):
        item_path = os.path.join(main_dir, item)
        if os.path.isdir(item_path):
            remove_files_with_tag(item_path, remove_tag)
        elif remove_tag in item:
            os.remove(item_path)



"""OpenFOAM module of the API for fluid dynamics."""
import pathlib
from typing import Optional

from inductiva import types
from inductiva.simulation import Simulator


class OpenFOAM(Simulator):
    """Class to invoke a generic DualSPHysics simulation on the API."""

    @property
    def api_method_name(self) -> str:
        return "fvm.openfoam.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        output_dir: Optional[types.Path] = None,
        openfoam_solver: str = "interFoam",
        n_cores: int = 1,
        track_logs: bool = False,
    ) -> pathlib.Path:
        """Run the simulation.

        Args:
            n_cores: Number of MPI cores to use for the simulation.
            openfoam_solver: specific solver to simulate with OpenFOAM.
                OpenFOAM contains lots of solvers inside of it, which are used
                to call the run simulation through terminal, e.g.,
                [isoFoam, sonicFoam, ...]. The default solver is interFoam.
            other arguments: See the documentation of the base class.
        """
        return super().run(
            input_dir,
            output_dir=output_dir,
            track_logs=track_logs,
            openfoam_solver=openfoam_solver,
            n_cores=n_cores,
        )

    def run_async(
        self,
        input_dir: types.Path,
        openfoam_solver: str = "interFoam",
        n_cores: int = 1,
    ) -> str:
        """Run the simulation asynchronously.
        
        Args:
            n_cores: Number of MPI cores to use for the simulation.
            openfoam_solver: specific solver to simulate with OpenFOAM.
                OpenFOAM contains lots of solvers inside of it, which are used
                to call the run simulation through terminal, e.g.,
                [isoFoam, sonicFoam, ...]. The default solver is interFoam.
            other arguments: See the documentation of the base class.
            """

        return super().run_async(input_dir,
                                 openfoam_solver=openfoam_solver,
                                 n_cores=n_cores)



"""Wind tunnel scenario to run an object over an air flow."""

from functools import singledispatchmethod
import os
import shutil
from typing import Optional, List

from absl import logging

from inductiva.types import Path
from inductiva.scenarios import Scenario
from inductiva.simulation import Simulator
from inductiva.fluids.simulators import OpenFOAM
from inductiva.utils.templates import (TEMPLATES_PATH,
                                       batch_replace_params_in_template)
from inductiva.utils.files import remove_files_with_tag

SCENARIO_TEMPLATE_DIR = os.path.join(TEMPLATES_PATH, "wind_tunnel")
OPENFOAM_TEMPLATE_INPUT_DIR = "openfoam"


class WindTunnel(Scenario):
    """Physical scenario of a configurable wind tunnel simulation.
    
    In this scenario, an object is inserted in a wind tunnel described by the
    user. The object is then subject to an air flow determined for which the
    direction and magnitude is defined by the user.
    """

    def __init__(self,
                 object_path: str,
                 flow_velocity: List[float],
                 domain: Optional[List[float]] = None):
        """Initializes a `WindTunnel` object.
        
        Args:
            object_path: Path to the object that is inserted in the wind tunnel.
            flow_velocity: Velocity of the air flow in m/s.
            domain: List containing the lower and upper boundary of the wind
                tunnel in each (x, y, z) direction. It is the natural
                description with the default OpenFOAM simulator.
        """

        self.object_path = object_path
        self.flow_velocity = flow_velocity

        if domain is None:
            logging.info("Using a default domain: [[-5, 15], [-4, 4], [0, 8]].")
            self.domain = [[-5, 15], [-4, 4], [0, 8]]
        else:
            self.domain = domain

    def simulate(self,
                 simulator: Simulator = OpenFOAM(),
                 output_dir: Optional[Path] = None,
                 simulation_time: float = 100,
                 output_time_step: float = 50,
                 n_cores: int = 1):
        """Simulates the wind tunnel scenario.
        
        Args:
            simulator: Simulator to use for the simulation.
            output_dir: Path to the directory where the simulation output
                is downloaded.
            simulation_time: Simulation time, in seconds.
            write_interval: Interval between simulation outputs, in seconds.
            n_cores: Number of cores to use for the simulation.
            """

        self.simulation_time = simulation_time
        self.output_time_step = output_time_step
        self.n_cores = n_cores

        output_path = super().simulate(
            simulator,
            output_dir=output_dir,
            n_cores=n_cores,
            openfoam_solver="simpleFoam",
        )

        return output_path

    @singledispatchmethod
    @classmethod
    def get_config_filename(cls, simulator: Simulator):  # pylint: disable=unused-argument
        raise ValueError(
            f"Simulator not supported for `{cls.__name__}` scenario.")

    @singledispatchmethod
    def gen_aux_files(self, simulator: Simulator, input_dir: str):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )

    @singledispatchmethod
    def gen_config(self, simulator: Simulator, input_dir: str):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )


@WindTunnel.get_config_filename.register
def _(self, simulator: OpenFOAM):  # pylint: disable=unused-argument
    pass


@WindTunnel.gen_aux_files.register
def _(self, simulator: OpenFOAM, input_dir):  # pylint: disable=unused-argument
    # The WindTunnel with OpenFOAM requires changing multiple files
    template_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                OPENFOAM_TEMPLATE_INPUT_DIR)

    # Copy all files from the template dir to the input directory
    for directory in os.listdir(template_dir):
        shutil.copytree(os.path.join(template_dir, directory),
                        os.path.join(input_dir, directory))

    # Remove all files that have .jinja in input_dir
    remove_files_with_tag(input_dir, ".jinja")


@WindTunnel.gen_config.register
def _(self, simulator: OpenFOAM, input_dir: str):  # pylint: disable=unused-argument
    """Generates the configuration files for OpenFOAM."""

    # The WindTunnel with OpenFOAM requires changing multiple files
    template_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                OPENFOAM_TEMPLATE_INPUT_DIR)

    # Add object path to its respective place
    object_input_dir_path = os.path.join(input_dir, "constant", "triSurface")
    os.mkdir(object_input_dir_path)
    shutil.copy(self.object_path,
                os.path.join(object_input_dir_path, "object.obj"))

    batch_replace_params_in_template(
        templates_dir=template_dir,
        template_filename_paths=[
            os.path.join("0", "include",
                         "initialConditions_template.openfoam.jinja"),
            os.path.join("system", "controlDict_template.openfoam.jinja"),
            os.path.join("system", "blockMeshDict_template.openfoam.jinja"),
            os.path.join("system", "decomposeParDict_template.openfoam.jinja")
        ],
        params={
            "flow_velocity": self.flow_velocity,
            "simulation_time": self.simulation_time,
            "output_time_step": self.output_time_step,
            "n_cores": self.n_cores,
            "domain": self.domain,
        },
        output_filename_paths=[
            os.path.join(input_dir, "0", "include", "initialConditions"),
            os.path.join(input_dir, "system", "controlDict"),
            os.path.join(input_dir, "system", "blockMeshDict"),
            os.path.join(input_dir, "system", "decomposeParDict")
        ])



"""Base class for scenarios."""

from abc import ABC
import tempfile
from typing import Optional

from inductiva.types import Path
from inductiva.simulation import Simulator
from inductiva.utils.files import resolve_path, get_timestamped_path
from inductiva.utils.misc import split_camel_case


class Scenario(ABC):
    """Base class for scenarios."""

    def simulate(
        self,
        simulator: Simulator,
        output_dir: Optional[Path] = None,
        **kwargs,
    ):
        """Simulates the scenario."""
        if output_dir is None:
            scenario_name_splitted = split_camel_case(self.__class__.__name__)
            output_dir_prefix = "-".join(scenario_name_splitted).lower()
            output_dir = get_timestamped_path(f"{output_dir_prefix}-output")

        output_dir = resolve_path(output_dir)

        with tempfile.TemporaryDirectory() as input_dir:

            self.gen_aux_files(simulator, input_dir)
            self.gen_config(simulator, input_dir)

            args = ()
            config_filename = self.get_config_filename(simulator)
            if config_filename:
                args += (config_filename,)

            output_path = simulator.run(
                input_dir,
                *args,
                output_dir=output_dir,
                **kwargs,
            )

        return output_path



"""Util functions to handle input and output data for API requests.

This module contains several functions related to packing and unpacking
of inputs and ouputs for Web API requests.
Additionally, it contains some global constant variables defining
configurations related to paths where certain files are expected to be.
"""
import os
import json
import pathlib
import zipfile
import tempfile
import shutil
import numpy as np
import scipy

from absl import logging

from .meta import is_tuple
from inductiva.types import Path

INPUT_FILENAME = "input.json"
OUTPUT_FILENAME = "output.json"
ARTIFACTS_DIRNAME = "artifacts"


def get_validate_request_params(original_params: dict,
                                type_annotations: dict) -> dict:
    """Convert original params to the params used for request validation.

    Convert a dictionary with the request params into the params that
    are used to validate the request.
    For instance, a numpy array is replaced by its shape,
    as that is what's necessary to validate if the user has permissions
    to execute the request.
    The original params are not modified.

    Args:
        original_params: Dict with the params that are passed into
            the request by the user.
        type_annotations: Dict with the type annotation of each param.

    Return:
        Returns a dictionary with the params for request validation.
    """
    params = {}

    for variable in original_params:
        param_type = type_annotations.get(variable, None)
        if param_type in (np.ndarray, scipy.sparse):
            params[variable] = {
                "shape": original_params[variable].shape,
            }
        elif param_type == pathlib.Path:
            # TODO: what kind of information do we want to send for validation
            # of a SplishSplash simulation?
            params[variable] = str(original_params[variable])
        else:
            params[variable] = original_params[variable]

    return params


def pack_param(name: str, value, param_type, dst_dir):
    """Pack a single parameter to the format expected by the Web API.

    Args:
        name: Name of the parameter.
        value: Original value of the parameter.
        param_type: Type annotation of the param.
        dst_dir: Directory in which to store files that may be required for
            some param types.

    Return:
        Returns a value that is passed to the API in a JSON file. For params
        that are sent as a file, the value is the path of the file relative
        to `dst_dir`.
    """
    if param_type == np.ndarray:
        param_filename = f"{name}.npy"
        param_fullpath = os.path.join(dst_dir, param_filename)
        np.save(param_fullpath, value)
        logging.debug("Stored %s to %s", name, param_fullpath)
        return param_filename

    if param_type == scipy.sparse:
        param_filename = f"{name}.npz"
        param_fullpath = os.path.join(dst_dir, param_filename)
        scipy.sparse.save_npz(param_fullpath, value)
        logging.debug("Stored %s to %s", name, param_fullpath)
        return param_filename

    if param_type == pathlib.Path:
        dst_dir_name = name
        dst_fullpath = os.path.join(dst_dir, dst_dir_name)

        shutil.copytree(value, dst_fullpath)

        logging.debug("Copied %s to %s", value, dst_fullpath)
        return str(dst_dir_name)

    return value


def pack_input(params, type_annotations, zip_name) -> str:
    """Pack all inputs into a zip file.

    Pack all input params and compress all files into a zip file.
    All required files are created in a temporary directory which is then
    compressed with "zip" format. The path of the resulting zip file is
    returned.

    Args:
        params: Dict with the params that are passed into
            the request by the user.
        type_annotations: Dict with the type annotation of each param.
            If a type annotation doesn't exist for a param, it is
            assumed that it is JSON encodable.
        zip_name: Name of the zip file to be created.

    Return:
        Returns a path to zip file with the compressed input. The zip will be
        located in the temporary directory of the OS (`/tmp` in linux).
    """
    with tempfile.TemporaryDirectory() as tmpdir_path:
        input_params = {}

        for variable in params:
            input_params[variable] = pack_param(
                name=variable,
                value=params[variable],
                param_type=type_annotations.get(variable, None),
                dst_dir=tmpdir_path,
            )

        # Write input dictionary with packed params to a JSON file
        input_json_path = os.path.join(tmpdir_path, INPUT_FILENAME)
        with open(input_json_path, "w", encoding="UTF-8") as fp:
            json.dump(input_params, fp)

        # Zip inputs
        zip_path = shutil.make_archive(
            os.path.join(tempfile.gettempdir(), zip_name), "zip", tmpdir_path)

        logging.debug("Compressed inputs to %s", zip_path)

    return zip_path


def unpack_value(value: str, var_type, output_dir: Path):
    """Unpack a single output value return by the Web API.

    Unpack a single output value, returning it as the correct type
    given the type annotation.

    Args:
        value: Output value to unpack. Can be a path relative to
            `output_dir` when the value is packed as a file.
        var_type: Type annotation of the value to unpack.
        output_dir: Directory where values packed as files are located.

    Return:
        Returns the unpacked value with the type defined by `var_type`.
    """
    if var_type == np.ndarray:
        return np.load(os.path.join(output_dir, value))

    if var_type == pathlib.Path:
        return pathlib.Path(os.path.join(output_dir, value))

    return type(value)


def unpack_output(result_list, output_dir: Path, return_type):
    """Transform outputs of a task executed in the API to the right objects.

    Args:
        result_list: List
        output_dir: Directory in which to store the extracted outputs.
        return_type: Type annotation of the method return type. If `return_type`
            is a tuple, the return is also a tuple with the all output values
            unpacked.

    Return:
        Returns the unpacked output of the method.
    """
    if return_type is None:
        return

    if return_type == pathlib.Path:
        return pathlib.Path(output_dir)

    if is_tuple(return_type):
        all_types = return_type.__args__

        return tuple(
            unpack_value(value, type, output_dir)
            for value, type in zip(result_list, all_types))

    return unpack_value(result_list[0], return_type, output_dir)


def extract_output(zip_path, output_dir):
    with zipfile.ZipFile(zip_path, "r") as zip_fp:
        output_json = zip_fp.read(OUTPUT_FILENAME)
        result_list = json.loads(output_json)

        extract_subdir_files(zip_fp, ARTIFACTS_DIRNAME, output_dir)
    return result_list


def extract_subdir_files(zip_fp: zipfile.ZipFile, dir_name: str,
                         output_dir: Path):
    """Util function to extract the contents of a directory in a ZIP archive.

    For instance, if a ZIP archive contains a directory called `dir_name`,
    the contents of that directory are extracted directly to `output_dir`.

    Args:
        zip_fp: ZipFile from which to extract the directory.
        dir_name: Name of the directory inside the ZIP archive.
        output_dir: Destination directory of the contents of `data_dir`.
    """
    for member in zip_fp.namelist():
        is_dir = not os.path.basename(member)

        if not member.startswith(dir_name) or is_dir:
            continue

        src_file = zip_fp.open(member)
        target_relative_path = pathlib.Path(member).relative_to(dir_name)
        target_path = os.path.join(output_dir, target_relative_path)

        os.makedirs(os.path.dirname(target_path), exist_ok=True)

        with open(target_path, "wb") as f:
            shutil.copyfileobj(src_file, f)


def zip_dir(dir_path, zip_name):
    """Compress a directory into a zip file."""
    zip_path = shutil.make_archive(
        os.path.join(tempfile.gettempdir(), zip_name), "zip", dir_path)

    logging.debug("Compressed inputs to %s", zip_path)

    return zip_path



"""Classes defining arbitrary geometrical shapes."""

from abc import abstractmethod
from dataclasses import dataclass, field
from typing import List, Tuple


# Shapes.
@dataclass
class BaseShape:
    """Base shape."""
    pass

    @abstractmethod
    def get_bounding_box(self) -> Tuple[List[float], List[float]]:
        """Returns the bounding box of the shape."""
        pass


@dataclass
class Base2DShape(BaseShape):
    """Base 2D shape."""
    position: List[float] = field(default_factory=lambda: [0, 0])


@dataclass
class Base3DShape(BaseShape):
    """Base 3D shape."""
    position: List[float] = field(default_factory=lambda: [0, 0, 0])


@dataclass
class Rectangle(Base2DShape):
    """Rectangular shape."""
    dimensions: List[float] = field(default_factory=lambda: [1, 1])

    def get_bounding_box(self) -> Tuple[List[float], List[float]]:
        """Returns the bounding box of the rectangle."""
        return (
            self.position,
            [self.position[i] + self.dimensions[i] for i in range(2)],
        )


@dataclass
class Circle(Base2DShape):
    """Circular shape."""
    radius: float = 1

    def get_bounding_box(self) -> Tuple[List[float], List[float]]:
        """Returns the bounding box of the circle."""
        return (
            [
                self.position[0] - self.radius,
                self.position[1] - self.radius,
            ],
            [
                self.position[0] + self.radius,
                self.position[1] + self.radius,
            ],
        )


@dataclass
class Cube(Base3DShape):
    """Cubic shape."""
    dimensions: List[float] = field(default_factory=lambda: [1, 1, 1])

    def get_bounding_box(self) -> Tuple[List[float], List[float]]:
        """Returns the bounding box of the cube."""
        return (
            self.position,
            [self.position[i] + self.dimensions[i] for i in range(3)],
        )


@dataclass
class Cylinder(Base3DShape):
    """Cylindrical shape."""
    radius: float = 0.5
    height: float = 1

    def get_bounding_box(self) -> Tuple[List[float], List[float]]:
        """Returns the bounding box of the cylinder."""
        return (
            [
                self.position[0] - self.radius,
                self.position[1] - self.radius,
                self.position[2],
            ],
            [
                self.position[0] + self.radius,
                self.position[1] + self.radius,
                self.position[2] + self.height,
            ],
        )



"""Methods that interact with the lower-level inductiva-web-api-client.

The relevant function for usage outside of this file is invoke_api().
Check the demos directory for examples on how it is used.
"""
import asyncio
import json
import os
import pathlib
import signal
import time
from contextlib import contextmanager
from typing import Any, Dict, Optional, Type
from urllib.parse import urlparse

import websockets
from absl import logging

import inductiva
from inductiva.client import ApiClient, ApiException, Configuration
from inductiva.client.apis.tags.tasks_api import TasksApi
from inductiva.client.models import (BodyUploadTaskInput, TaskRequest,
                                     TaskStatus)
from inductiva.exceptions import RemoteExecutionError
from inductiva.types import Path
from inductiva.utils.data import (extract_output, get_validate_request_params,
                                  pack_input, unpack_output)
from inductiva.utils.meta import get_method_name, get_type_annotations


def validate_api_key(api_key: str) -> Configuration:
    """Validates the API key and returns API configuration"""
    if inductiva.api_key is None:
        raise ValueError(
            "No API Key specified. "
            "Set it in the code with \"inductiva.api_key = <YOUR_SECRET_KEY>\""
            " or set the INDUCTIVA_API_KEY environment variable.")

    api_config = Configuration(host=inductiva.api_url)
    api_config.api_key["APIKeyHeader"] = api_key

    return api_config


def submit_request(api_instance: TasksApi, request: TaskRequest) -> TaskStatus:
    """Submits a task request to the API.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
    Return:
        Returns the body of the HTTP response.
        Contains two fields, "id" and "status".
    """

    try:
        api_response = api_instance.submit_task(body=request)
    except ApiException as e:
        logging.exception("Exception when calling TasksApi->submit_task: %s", e)
        raise e

    logging.debug("Request status: %s", api_response.body["status"])

    return api_response.body


def upload_input(api_instance: TasksApi, task_id, original_params,
                 type_annotations):
    """Uploads the inputs of a given task to the API.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task.
        original_params: Params of the request passed by the user.
        type_annotations: Annotations of the params' types.
    """
    input_zip_path = pack_input(
        params=original_params,
        type_annotations=type_annotations,
        zip_name=task_id,
    )

    logging.info("Uploading input files...")
    try:
        with open(input_zip_path, "rb") as zip_fp:
            _ = api_instance.upload_task_input(
                path_params={"task_id": task_id},
                body=BodyUploadTaskInput(file=zip_fp),
            )
    except ApiException as e:
        logging.exception(
            "Exception when calling TasksApi->upload_task_inputs: %s", e)
        raise e

    os.remove(input_zip_path)


def download_output(api_instance: TasksApi,
                    task_id,
                    output_dir: Optional[Path] = None):
    """Downloads the output of a given task from the API.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task.

    Return:
        Downloads and extracts the data to the client.
    """
    logging.info("Downloading output...")
    try:
        api_response = api_instance.download_task_output(
            path_params={"task_id": task_id},
            stream=True,
        )
        logging.info("Output downloaded.")
    except ApiException as e:
        raise e

    logging.debug("Downloaded output to %s", api_response.body.name)

    if output_dir is None:
        output_dir = os.path.join(inductiva.output_dir, task_id)

    logging.info("Extracting output ZIP file to \"%s\"...", output_dir)
    result_list = extract_output(api_response.body.name, output_dir)
    logging.info("Output extracted.")

    return result_list


def block_until_finish(api_instance: TasksApi, task_id: str) -> str:
    """Block until a task executing remotely finishes execution.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task to wait for.

    Returns:
        Returns info related to the task, containing two fields,
        "id" and "status".
    """
    logging.debug("Blocking until task is finished ...")
    return block_until_status_is(api_instance, task_id, {"success", "failed"})


def kill_task(api_instance: TasksApi, task_id: str):
    """Kill a task that is executing remotely.

    The function sends a kill request to the API.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task to kill.
   """
    logging.debug("Sending kill task request ...")
    api_instance.kill_task(path_params={"task_id": task_id},)
    logging.info("Task terminated.")


def get_task_status(api_instance: TasksApi, task_id: str) -> TaskStatus:
    """Check the status of a task."""

    api_response = api_instance.get_task_status(
        path_params={"task_id": task_id})

    status = api_response.body["status"]

    return status


def block_until_status_is(api_instance: TasksApi,
                          task_id,
                          desired_status,
                          sleep_secs=0.5):
    """Block until the status of a task becomes the desired status.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task to wait for.
        desired_status: Task status to wait for.
        sleep_secs: Polling interval.

    Returns:
        Returns info related to the task, containing two fields,
    """
    prev_status = None

    while True:
        try:
            status = get_task_status(api_instance, task_id)
            logging.debug("Task status is %s", status)
        except ApiException as e:
            raise e

        if status != prev_status:
            if status == "submitted":
                logging.info("Waiting for resources...")

            prev_status = status

        # If status reaches the desired status, then stop polling
        if status in desired_status:
            break

        time.sleep(sleep_secs)

    return status


@contextmanager
def blocking_task_context(api_instance: TasksApi, task_id):
    """Context to handle execution of a blocking task.

    The context handles exceptions and the SIGINT signal, issuing a request
    to the API to kill the executing task.
    Info on the implementation:
    - https://docs.python.org/3/library/contextlib.html

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task being executed.
    """
    # Other imported modules can make changes to the SIGINT handler, so we set
    # the default int handler to make sure that KeyboardInterrupt is raised
    # if the user presses Ctrl+C.
    original_sig = signal.signal(signal.SIGINT, signal.default_int_handler)

    try:
        yield None
    except Exception as err:
        logging.info("Caught exception: terminating blocking task...")
        kill_task(api_instance, task_id)
        raise err
    except KeyboardInterrupt as err:
        logging.info("Caught SIGINT: terminating blocking task...")
        kill_task(api_instance, task_id)
        raise err
    finally:
        # Reset original SIGINT handler
        signal.signal(signal.SIGINT, original_sig)


def invoke_api_from_fn_ptr(
    params,
    function_ptr,
    output_dir: Optional[Path] = None,
    track_logs: bool = False,
):
    """Perform a task remotely defined by a function pointer."""
    type_annotations = get_type_annotations(function_ptr)
    method_name = get_method_name(function_ptr)
    return invoke_api(
        method_name,
        params,
        output_dir=output_dir,
        type_annotations=type_annotations,
        return_type=type_annotations["return"],
        log_remote_execution=track_logs,
    )


def submit_task(api_instance, task_request, params, type_annotations):
    """Submit a task and send input files to the API."""

    task = submit_request(
        api_instance=api_instance,
        request=task_request,
    )

    task_id = task["id"]
    logging.info("This simulation has the task_id: %s", task_id)

    with blocking_task_context(api_instance, task_id):
        if task["status"] == "pending-input":

            upload_input(
                api_instance=api_instance,
                original_params=params,
                task_id=task_id,
                type_annotations=type_annotations,
            )

            logging.info("Request submitted.")

    return task_id


def invoke_api(method_name: str,
               params,
               type_annotations: Dict[Any, Type],
               output_dir: Optional[Path] = None,
               return_type: Type = pathlib.Path,
               log_remote_execution: bool = False):
    """Perform a task remotely via Inductiva's Web API.

    Currently, the implementation handles the whole flow of the task execution,
    and blocks until the task finishes execution.
    The flow is summarized as follows:
        1. Transform request params into the params used to
        validate permission to execute the request.
        2. Submit task via the "POST task/submit" endpoint.
            Note: HTTP status code 400 informs the requested method is invalid,
                and HTTP status code 403 informs that the user is not authorized
                to post such request.
        3. If the status returned by the previous HTTP request is
            "pending-input", ZIP inputs and send them via
            "POST task/{task_id}/input".
        4. Block execution, polling using the "GET task/{task_id}/status"
            endpoint until the returned status is "success".
        5. Download output ZIP with "GET task/{task_id}/output".
        6. Unpack output ZIP and return correct value based on the
            type annotations of `function_ptr`.

    Args:
        request: Request sent to the API for validation.
        input_dir: Directory containing the input files to be uploaded.
        output_dir: Directory where to place the output files.
        return_type: Type of the return value of the task, for unpacking.

    Return:
        Returns the output of the task.
    """
    api_config = validate_api_key(inductiva.api_key)

    request_params = get_validate_request_params(
        original_params=params,
        type_annotations=type_annotations,
    )

    task_request = TaskRequest(
        method=method_name,
        params=request_params,
    )

    with ApiClient(api_config) as client:
        api_instance = TasksApi(client)

        task_id = submit_task(api_instance, task_request, params,
                              type_annotations)

        block_until_status_is(
            api_instance=api_instance,
            task_id=task_id,
            desired_status={"started"},
        )
        logging.info("An executer has picked up the request")
        logging.info("The requested task is being executed remotely")

        # While the task is executing, use a context manager that kills the
        # task if some exception or SIGINT is caught.
        with blocking_task_context(api_instance, task_id):
            if log_remote_execution:
                status = asyncio.run(follow_task(task_id))
            else:
                status = block_until_finish(
                    api_instance=api_instance,
                    task_id=task_id,
                )

            if status == "success":
                logging.info("Task executed successfuly.")
            else:
                logging.info("Task failed.")

        result_list = download_output(api_instance, task_id, output_dir)

        if status == "failed":
            raise RemoteExecutionError(f"""Remote execution failed.
        Find more details in: \"{os.path.abspath(output_dir)}\".""")

    return unpack_output(result_list, output_dir, return_type)


def invoke_async_api(method_name: str, params, type_annotations: Dict[Any,
                                                                      Type]):
    """Perform a task asyc and remotely via Inductiva's Web API.

    Submits a simulation async to the API and returns the task id.
    The flow is similar to invoke_api, but it does not block until the task
    is finished. 
    The flow is summarized as follows:
        1. Transform request params into the params used to
        validate permission to execute the request.
        2. Submit task via the "POST task/submit" endpoint.
            Note: HTTP status code 400 informs the requested method is invalid,
                and HTTP status code 403 informs that the user is not authorized
                to post such request.
        3. If the status returned by the previous HTTP request is
            "pending-input", ZIP inputs and send them via
            "POST task/{task_id}/input".
        4. Return task_id and leaves the simulation on the queue until resources
            become available.

    Args:
        request: Request sent to the API for validation.
        input_dir: Directory containing the input files to be uploaded.

    Return:
        Returns the task id.
    """

    api_config = validate_api_key(inductiva.api_key)

    request_params = get_validate_request_params(
        original_params=params,
        type_annotations=type_annotations,
    )

    task_request = TaskRequest(
        method=method_name,
        params=request_params,
    )

    with ApiClient(api_config) as client:
        api_instance = TasksApi(client)

        task_id = submit_task(api_instance=api_instance,
                              task_request=task_request,
                              params=params,
                              type_annotations=type_annotations)

    return task_id


def run_simulation(
    api_method_name: str,
    input_dir: pathlib.Path,
    output_dir: pathlib.Path,
    params: Dict[str, Any],
    log_remote_execution: bool = False,
) -> pathlib.Path:
    """Run a simulation synchronously via Inductiva Web API."""

    params = {
        "sim_dir": input_dir,
        **params,
    }
    type_annotations = {
        "sim_dir": pathlib.Path,
    }

    result = invoke_api(
        api_method_name,
        params,
        type_annotations,
        output_dir=output_dir,
        return_type=pathlib.Path,
        log_remote_execution=log_remote_execution,
    )

    if not isinstance(result, pathlib.Path):
        raise RuntimeError(f"Expected result to be a Path, got {type(result)}")

    return result


def run_async_simulation(
    api_method_name: str,
    input_dir: pathlib.Path,
    params: Dict[str, Any],
) -> str:
    """Run a simulation asynchronously via Inductiva Web API."""

    params = {
        "sim_dir": input_dir,
        **params,
    }
    type_annotations = {
        "sim_dir": pathlib.Path,
    }

    result = invoke_async_api(
        api_method_name,
        params,
        type_annotations,
    )

    if not isinstance(result, str):
        raise RuntimeError(
            f"Expected result to be a string with task_id, got {type(result)}")

    return result


async def follow_task(task_id: str) -> str:
    parsed_url = urlparse(inductiva.api_url)
    url = \
        f"ws://{parsed_url.hostname}:{parsed_url.port}/" \
        f"task/{task_id}/logs/tail"

    logging.debug("Following task logs at %s", url)
    async with websockets.connect(url) as ws:
        while True:
            message = await ws.recv()
            parsed_message = json.loads(message)

            if parsed_message["type"] == "update":
                status = parsed_message["content"]
                logging.debug("Task status update: %s", status)

                return status
            else:
                new_line = parsed_message["content"]
                print(new_line, end="")



"""Methods to interact with the tasks submitted to the API."""
import pathlib
from typing import Optional, Type

from absl import logging

import inductiva
from inductiva import api
from inductiva.types import Path
from inductiva.client import ApiClient
from inductiva.client.apis.tags.tasks_api import TasksApi
from inductiva.utils.data import (unpack_output)


def get_task_status(task_id: str):
    """Get the status of a simulation."""

    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = TasksApi(client)

        status = api.get_task_status(api_instance, task_id)
    logging.info("Task status: %s", status)

    return status


def fetch_task_output(task_id: str,
                      output_dir: Optional[Path] = None,
                      return_type: Type = pathlib.Path) -> Type:
    """Fetch the results of a simulation."""
    logging.info("Fetching the simulation output for task %s", task_id)
    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = TasksApi(client)

        status = api.get_task_status(api_instance, task_id)

        logging.info(status)
        if status == "started":
            logging.info("The task is still being executed.")
            return_type = None
        elif status == "submitted":
            logging.info("The task is waiting for resources...")
            return_type = None
        else:
            result_list = api.download_output(api_instance, task_id, output_dir)

            return unpack_output(result_list, output_dir, return_type)



"""Visualization utilities."""

import os
import tempfile
from typing import Dict, List, Optional

from absl import logging

import imageio
import pyvista as pv
import matplotlib
import matplotlib.colors as clr
import matplotlib.pyplot as plt
from tqdm import tqdm
import xarray as xr

from inductiva.utils import files

MPL_CONFIG_PARAMS = {
    "font.size": 14,
    "axes.titlesize": "medium",
    "figure.dpi": 100,
}


def create_movie_from_frames(frames_dir: str,
                             movie_path: str,
                             fps: int = 10) -> None:
    """Creates movie from a series of png image files.

    The order of the png image file names determines the order with which they
    are rendered in the movie. For example, image 'frame-001.png' will appear
    before 'frame-002.png'.

    Args:
        frames_dir: Directory where the png files are stored.
        movie_path: Path to the movie to be created.
        fps: Number of frames per second to use in the movie.
    """

    writer = imageio.get_writer(movie_path, fps=fps)

    frames_list = [
        file for file in os.listdir(frames_dir) if file.endswith(".png")
    ]
    frames_list.sort()

    for file_name in frames_list:
        frame_path = os.path.join(frames_dir, file_name)
        writer.append_data(imageio.imread(frame_path))

    writer.close()


def create_movie_from_vtk(vtk_output_dir: str,
                          movie_path: str,
                          virtual_display: bool = True,
                          scalars: str = None,
                          scalar_limits: Optional[List[float]] = None,
                          camera=None,
                          color: str = "blue",
                          cmap: str = None,
                          fps: int = 10) -> None:
    """Creates movie from a series of vtk files.

    The order of the vtk file name determines the order with which they
    are rendered in the movie. For example, vtk file 'frame_001.vtk' will
    appear before 'frame_002.vtk'.

    Args:
        vtk_output_dir: Directory containing the vtk files.
        movie_path: Path to save the movie.
        virtual_display: Whether to use a virtual display to render
            the movie.
        scalar: Scalars used to “color” the mesh. Accepts a string name
            of an array that is present on the mesh or an array equal to
            the number of cells or the number of points in the mesh.
            Array should be sized as a single vector.
        scalar_bounds: Color bar range for scalars. Defaults to minimum
            and maximum of scalars array. Example: [-1, 2].
        objects: Object of pyvista.PolyData type describing the domain or
            an object inside.
        camera: Camera description must be one of the following:
          - List of three tuples describing the position, focal-point
          and view-up: [(2.0, 5.0, 13.0), (0.0, 0.0, 0.0), (-0.7, -0.5, 0.3)]
          - List with a view-vector: [-1.0, 2.0, -5.0]
          - A string with the plane orthogonal to the view direction: 'xy'.
          https://docs.pyvista.org/api/plotting/_autosummary/pyvista.CameraPosition.html
        color: Color of the points used to represent particles. Default: "blue".
        cmap: string with the name of the matplotlib colormap to use
            when mapping the scalars. See available Matplotlib colormaps.
        fps: Number of frames per second to use in the movie. Renders only a
            subset of the vtk files to create the movie. This is done for
            speed purposes.
            Default: 10.
    """
    if virtual_display:
        pv.start_xvfb()

    pv.global_theme.background = "white"
    vtk_output_dir = files.resolve_path(vtk_output_dir)
    vtk_files = files.get_sorted_files(vtk_output_dir, ".vtk")

    with tempfile.TemporaryDirectory() as tmp_dir:
        logging.info("Creating movie frames...")
        for index, frame_file in enumerate(vtk_files):
            if index % int(round(60 / fps)) == 0:
                frame_path = os.path.join(vtk_output_dir, frame_file)
                image_frame_path = os.path.join(
                    tmp_dir, "frame_" + str(index).zfill(5) + ".png")

                create_frame_from_vtk(frame_path,
                                      image_frame_path=image_frame_path,
                                      camera=camera,
                                      scalars=scalars,
                                      scalar_limits=scalar_limits,
                                      color=color,
                                      cmap=cmap)

        logging.info("Creating movie '%s'.", movie_path)
        create_movie_from_frames(frames_dir=tmp_dir,
                                 movie_path=movie_path,
                                 fps=fps)


def create_frame_from_vtk(frame_path: str,
                          image_frame_path: str,
                          scalars: str = None,
                          scalar_limits: Optional[List[float]] = None,
                          camera=None,
                          color: str = None,
                          cmap: str = None):
    """Render a .png image from a vtk file."""

    frame = pv.read(frame_path)

    frame.plot(off_screen=True,
               screenshot=image_frame_path,
               cpos=camera,
               scalars=scalars,
               clim=scalar_limits,
               render_points_as_spheres=True,
               color=color,
               cmap=cmap)


def create_2d_scatter_plot(
    dataset: xr.Dataset,
    x_var: str,
    y_var: str,
    image_path: str,
    marker_size: float = 20,
    alpha: float = 1,
    color: Optional[str] = None,
    color_var: Optional[str] = None,
    x_limits: Optional[List[float]] = None,
    y_limits: Optional[List[float]] = None,
    color_limits: Optional[List[float]] = None,
    color_map: Optional[str] = "viridis",
    aspect: Optional[str] = None,
    matplotlib_config_params: Optional[Dict] = None,
):
    """Creates a 2d scatter plot of a dataset."""

    if matplotlib_config_params is None:
        matplotlib_config_params = MPL_CONFIG_PARAMS

    with matplotlib.rc_context(rc=matplotlib_config_params):
        fig = plt.figure()
        ax = fig.add_subplot()

        color_norm = None
        if color_var is not None and color_limits is not None:
            color_norm = clr.Normalize(*color_limits)

        dataset.plot.scatter(
            ax=ax,
            x=x_var,
            y=y_var,
            hue=color_var,
            s=marker_size,
            alpha=alpha,
            c=color,
            xlim=x_limits,
            ylim=y_limits,
            norm=color_norm,
            cmap=color_map,
        )

        if aspect is not None:
            ax.set(aspect=aspect)

        fig.tight_layout()
        fig.savefig(image_path)
        plt.close(fig)


def create_3d_scatter_plot(
    dataset: xr.Dataset,
    x_var: str,
    y_var: str,
    z_var: str,
    image_path: str,
    marker_size: float = 20,
    alpha: float = 1,
    color: Optional[str] = None,
    color_var: Optional[str] = None,
    x_limits: Optional[List[float]] = None,
    y_limits: Optional[List[float]] = None,
    z_limits: Optional[List[float]] = None,
    color_limits: Optional[List[float]] = None,
    color_map: Optional[str] = "viridis",
    matplotlib_config_params: Optional[Dict] = None,
):
    """Creates a 3d scatter plot of a dataset."""

    if matplotlib_config_params is None:
        matplotlib_config_params = MPL_CONFIG_PARAMS

    with matplotlib.rc_context(rc=matplotlib_config_params):
        fig = plt.figure()
        ax = fig.add_subplot(projection="3d")

        color_norm = None
        if color_var is not None and color_limits is not None:
            color_norm = clr.Normalize(*color_limits)

        dataset.plot.scatter(
            ax=ax,
            x=x_var,
            y=z_var,
            z=y_var,
            s=marker_size,
            alpha=alpha,
            c=color,
            hue=color_var,
            xlim=x_limits,
            ylim=z_limits,
            norm=color_norm,
            cmap=color_map,
        )

        ax.set_zlim(y_limits)

        fig.tight_layout()
        fig.savefig(image_path)
        plt.close(fig)


def create_2d_scatter_plot_movie(
    dataset: xr.Dataset,
    iter_var: str,
    x_var: str,
    y_var: str,
    movie_path: str,
    movie_fps: int = 10,
    marker_size: float = 20,
    alpha: float = 1,
    color: Optional[str] = None,
    color_var: Optional[str] = None,
    x_limits: Optional[List[float]] = None,
    y_limits: Optional[List[float]] = None,
    color_limits: Optional[List[float]] = None,
    color_map: Optional[str] = "viridis",
    aspect: Optional[str] = None,
    matplotlib_config_params: Optional[Dict] = None,
):
    """Creates a movie of the dataset with 2d scatter plots."""

    with tempfile.TemporaryDirectory() as tmp_dir:
        logging.info("Creating movie frames...")
        var_iterator = dataset.groupby(iter_var)
        for i, (_, dataset_i) in tqdm(enumerate(var_iterator),
                                      total=len(var_iterator)):
            create_2d_scatter_plot(
                dataset=dataset_i,
                x_var=x_var,
                y_var=y_var,
                marker_size=marker_size,
                alpha=alpha,
                color=color,
                color_var=color_var,
                x_limits=x_limits,
                y_limits=y_limits,
                color_limits=color_limits,
                color_map=color_map,
                aspect=aspect,
                image_path=os.path.join(tmp_dir, f"frame_{i:06d}.png"),
                matplotlib_config_params=matplotlib_config_params,
            )

        logging.info("Finished creating movie frames.")

        logging.info("Creating movie '%s'.", movie_path)
        create_movie_from_frames(frames_dir=tmp_dir,
                                 movie_path=movie_path,
                                 fps=movie_fps)


def create_3d_scatter_plot_movie(
    dataset: xr.Dataset,
    iter_var: str,
    x_var: str,
    y_var: str,
    z_var: str,
    movie_path: str,
    movie_fps: int = 10,
    marker_size: float = 20,
    alpha: float = 1,
    color: Optional[str] = None,
    color_var: Optional[str] = None,
    x_limits: Optional[List[float]] = None,
    y_limits: Optional[List[float]] = None,
    z_limits: Optional[List[float]] = None,
    color_limits: Optional[List[float]] = None,
    color_map: Optional[str] = "viridis",
    matplotlib_config_params: Optional[Dict] = None,
):
    """Creates a movie of the dataset with 3d scatter plots."""

    with tempfile.TemporaryDirectory() as tmp_dir:
        logging.info("Creating movie frames...")
        var_iterator = dataset.groupby(iter_var)
        for i, (_, dataset_i) in tqdm(enumerate(var_iterator),
                                      total=len(var_iterator)):
            create_3d_scatter_plot(
                dataset=dataset_i,
                x_var=x_var,
                y_var=y_var,
                z_var=z_var,
                marker_size=marker_size,
                alpha=alpha,
                color=color,
                color_var=color_var,
                x_limits=x_limits,
                y_limits=y_limits,
                z_limits=z_limits,
                color_limits=color_limits,
                color_map=color_map,
                image_path=os.path.join(tmp_dir, f"frame_{i:06d}.png"),
                matplotlib_config_params=matplotlib_config_params,
            )

        logging.info("Finished creating movie frames.")

        logging.info("Creating movie '%s'.", movie_path)
        create_movie_from_frames(frames_dir=tmp_dir,
                                 movie_path=movie_path,
                                 fps=movie_fps)


def create_color_plot(
    data_array: xr.DataArray,
    image_path: str,
    x_limits: Optional[List[float]] = None,
    y_limits: Optional[List[float]] = None,
    color_limits: Optional[List[float]] = None,
    color_map: Optional[str] = "viridis",
    aspect: Optional[str] = None,
    matplotlib_config_params: Optional[Dict] = None,
):
    """Creates a color plot of a data array."""

    if matplotlib_config_params is None:
        matplotlib_config_params = MPL_CONFIG_PARAMS

    with matplotlib.rc_context(rc=matplotlib_config_params):
        fig = plt.figure()
        ax = fig.add_subplot()

        color_norm = None
        if color_limits is not None:
            color_norm = clr.Normalize(*color_limits)

        data_array.plot.pcolormesh(
            ax=ax,
            xlim=x_limits,
            ylim=y_limits,
            norm=color_norm,
            cmap=color_map,
        )

        if aspect is not None:
            ax.set(aspect=aspect)

        fig.tight_layout()
        fig.savefig(image_path)
        plt.close(fig)


def create_color_plot_movie(
    data_array: xr.DataArray,
    iter_var: str,
    movie_path: str,
    movie_fps: int = 10,
    x_limits: Optional[List[float]] = None,
    y_limits: Optional[List[float]] = None,
    color_limits: Optional[List[float]] = None,
    color_map: Optional[str] = "viridis",
    aspect: Optional[str] = None,
    matplotlib_config_params: Optional[Dict] = None,
):
    """Creates a movie of the data array with color plots."""

    with tempfile.TemporaryDirectory() as tmp_dir:
        logging.info("Creating movie frames...")
        var_iterator = data_array.groupby(iter_var)
        for i, (_, data_array_i) in tqdm(enumerate(var_iterator),
                                         total=len(var_iterator)):
            create_color_plot(
                data_array=data_array_i,
                x_limits=x_limits,
                y_limits=y_limits,
                color_limits=color_limits,
                color_map=color_map,
                aspect=aspect,
                image_path=os.path.join(tmp_dir, f"frame_{i:06d}.png"),
                matplotlib_config_params=matplotlib_config_params,
            )

        logging.info("Finished creating movie frames.")

        logging.info("Creating movie '%s'.", movie_path)
        create_movie_from_frames(frames_dir=tmp_dir,
                                 movie_path=movie_path,
                                 fps=movie_fps)



"""SWASH module of the API."""
import pathlib
from typing import Optional
from inductiva.simulation import Simulator
from inductiva import types


class SWASH(Simulator):
    """Class to invoke a generic SWASH simulation on the API."""

    @property
    def api_method_name(self) -> str:
        return "sw.swash.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        sim_config_filename: str,
        output_dir: Optional[types.Path] = None,
        n_cores: int = 1,
        track_logs: bool = False,
    ) -> pathlib.Path:
        """Run the simulation.

        Args:
            n_cores: Number of MPI cores to use for the simulation.
            sim_config_filename: Name of the simulation configuration file.
            other arguments: See the documentation of the base class.
        """
        return super().run(input_dir,
                           track_logs=track_logs,
                           input_filename=sim_config_filename,
                           n_cores=n_cores,
                           output_dir=output_dir)

    def run_async(
        self,
        input_dir: types.Path,
        sim_config_filename: str,
        n_cores: int = 1,
    ) -> str:
        """Run the simulation asynchronously.
        
        Args:
            sim_config_filename: Name of the simulation configuration file.
            n_cores: Number of MPI cores to use for the simulation.
            """

        return super().run_async(input_dir,
                                 n_cores=n_cores,
                                 input_filename=sim_config_filename)



"""Functions that use the API to perform basic math operations.

These functions are meant for proof-of-concept and testing.
For example usage, refer to the demos directory of this repository.
"""
import numpy as np
from inductiva.api import invoke_api_from_fn_ptr
# pylint: disable=unused-argument, disable=redefined-builtin


def sum(a: float, b: float) -> float:
    return invoke_api_from_fn_ptr(locals(), sum)


def matmul(m: np.ndarray, n: np.ndarray) -> np.ndarray:
    return invoke_api_from_fn_ptr(locals(), matmul)


# pylint: enable=unused-argument, enable=redefined-builtin



"""Base class for low-level simulators."""
from abc import ABC, abstractmethod
import pathlib
from typing import Optional

from inductiva import api
from inductiva import types
from inductiva.utils import files


class Simulator(ABC):
    """Base simulator class."""

    @property
    @abstractmethod
    def api_method_name(self) -> str:
        pass

    def run(
        self,
        input_dir: types.Path,
        *_args,
        output_dir: Optional[types.Path] = None,
        track_logs: bool = False,
        **kwargs,
    ) -> pathlib.Path:
        """Run the simulation.

        Args:
            input_dir: Path to the directory containing the input files.
            _args: Unused in this method, but defined to allow for more
                non-default arguments in method override in subclasses.
            output_dir: Path to the directory where the output files will be
                stored. If not provided, a timestamped directory will be
                created with the same name as the input directory appended
                with "-output".
            track_logs: If True, the logs of the remote execution will be
                streamed to the console.
            **kwargs: Additional keyword arguments to be passed to the
                simulation API method.
        """
        input_dir = files.resolve_path(input_dir)
        if not input_dir.is_dir():
            raise ValueError(
                f"The provided path (\"{input_dir}\") is not a directory.")

        if output_dir is None:
            output_dir = input_dir.with_name(f"{input_dir.name}-output")
            output_dir = files.get_timestamped_path(output_dir)
        else:
            output_dir = files.resolve_path(output_dir)

        return api.run_simulation(
            self.api_method_name,
            input_dir,
            output_dir,
            params=kwargs,
            log_remote_execution=track_logs,
        )

    def run_async(
        self,
        input_dir: types.Path,
        *_args,
        **kwargs,
    ) -> str:
        """Run the simulation asynchronously.

        Args:
            input_dir: Path to the directory containing the input files.
            _args: Unused in this method, but defined to allow for more
                non-default arguments in method override in subclasses.
            **kwargs: Additional keyword arguments to be passed to the
                simulation API method.
        """
        input_dir = files.resolve_path(input_dir)
        if not input_dir.is_dir():
            raise ValueError(
                f"The provided path (\"{input_dir}\") is not a directory.")

        return api.run_async_simulation(
            self.api_method_name,
            input_dir,
            params=kwargs,
        )



# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from inductiva.client.apis.tag_to_api import tag_to_api

import enum


class TagValues(str, enum.Enum):
    ADMIN = "admin"
    TASKS = "tasks"



"""Post-processing utilities for SPlisHSPlasH."""

import os
import pathlib

from absl import logging

import numpy as np
from tqdm import tqdm
import vtk
from vtk.util.numpy_support import vtk_to_numpy as _vtk_data_to_numpy
import xarray as xr

from inductiva.utils.files import get_sorted_files


def convert_vtk_data_dir_to_netcdf(
    data_dir: str,
    output_time_step: float,
    netcdf_data_dir: str,
):
    """Converts simulation output files to netcdf format.

    Args:
        data_dir: Data directory.
        output_time_step: Time step between output files, in seconds.
        netcdf_data_dir: Directory to store files in netcdf format.
    """

    files = get_sorted_files(data_dir, ".vtk")

    if not os.path.exists(netcdf_data_dir):
        os.makedirs(netcdf_data_dir)

    logging.info("Converting vtk files to netcdf format...")
    for file_key, file in tqdm(enumerate(files), total=len(files)):
        time = file_key * output_time_step
        xr_dataset = read_vtk_file_to_xr_dataset(file.path, time)
        file_stem = pathlib.Path(file.path).stem
        xr_dataset.to_netcdf(os.path.join(netcdf_data_dir, f"{file_stem}.nc"))


def read_vtk_file_to_xr_dataset(file_path: str, time: float) -> xr.Dataset:
    """Reads a single simulation output file to an xarray Dataset.

    Args:
        file_path: File path.
        time: Time instant, in seconds, associated with the data in the file.
    """

    if not os.path.exists(file_path):
        raise FileExistsError(f"File '{file_path}' not found.")

    if pathlib.Path(file_path).suffix != ".vtk":
        raise IOError(f"File '{file_path}' does not have .vtk extension.")

    time_var = xr.DataArray(np.asarray([time]),
                            attrs={
                                "units": "s",
                                "long_name": "$t$"
                            })

    data_vars = {}

    # Create vtk file reader.
    # SPlisHSPlasH saves particle data as an unstructured grid, so the
    # reader must be of that type.
    vtk_reader = vtk.vtkUnstructuredGridReader()

    # Set the file path in the reader.
    vtk_reader.SetFileName(file_path)
    vtk_reader.Update()

    # Read the file output, i.e. the type, size, etc. of its contents.
    vtk_output = vtk_reader.GetOutput()

    num_particles = vtk_output.GetNumberOfPoints()
    particle_idx_var = xr.DataArray(np.arange(num_particles),
                                    attrs={"long_name": "particle index"})

    # Read position data.
    position_data = _vtk_data_to_numpy(vtk_output.GetPoints().GetData())

    data_vars["x"] = xr.DataArray(
        position_data[np.newaxis, :, 0],
        dims=["time", "particle_idx"],
        coords=[time_var, particle_idx_var],
        attrs={
            "units": "m",
            "long_name": "$x$"
        },
    )

    data_vars["y"] = xr.DataArray(
        position_data[np.newaxis, :, 1],
        dims=["time", "particle_idx"],
        coords=[time_var, particle_idx_var],
        attrs={
            "units": "m",
            "long_name": "$y$"
        },
    )

    data_vars["z"] = xr.DataArray(
        position_data[np.newaxis, :, 2],
        dims=["time", "particle_idx"],
        coords=[time_var, particle_idx_var],
        attrs={
            "units": "m",
            "long_name": "$z$"
        },
    )

    # Read vtk point data, i.e. data associated with each particle other
    # than position.
    vtk_point_data = vtk_output.GetPointData()

    # Get names of all arrays of point data to validate data reading.
    vtk_point_data_array_names = [
        vtk_point_data.GetArray(idx).GetName()
        for idx in range(vtk_point_data.GetNumberOfArrays())
    ]

    if "velocity" in vtk_point_data_array_names:

        # Read velocity data.
        velocity_data = _vtk_data_to_numpy(vtk_point_data.GetArray("velocity"))

        data_vars["velocity_x"] = xr.DataArray(
            velocity_data[np.newaxis, :, 0],
            dims=["time", "particle_idx"],
            coords=[time_var, particle_idx_var],
            attrs={
                "units": "m/s",
                "long_name": "$v_x$"
            },
        )

        data_vars["velocity_y"] = xr.DataArray(
            velocity_data[np.newaxis, :, 1],
            dims=["time", "particle_idx"],
            coords=[time_var, particle_idx_var],
            attrs={
                "units": "m/s",
                "long_name": "$v_y$"
            },
        )

        data_vars["velocity_z"] = xr.DataArray(
            velocity_data[np.newaxis, :, 2],
            dims=["time", "particle_idx"],
            coords=[time_var, particle_idx_var],
            attrs={
                "units": "m/s",
                "long_name": "$v_z$"
            },
        )

    if "density" in vtk_point_data_array_names:

        # Read density data.
        density_data = _vtk_data_to_numpy(vtk_point_data.GetArray("density"))

        data_vars["density"] = xr.DataArray(
            density_data[np.newaxis, :],
            dims=["time", "particle_idx"],
            coords=[time_var, particle_idx_var],
            attrs={
                "units": "kg/m$^3$",
                "long_name": "$\\rho$"
            },
        )

    return xr.Dataset(data_vars)



"""Functions to explore design space of simulations."""
import os
from typing import Any, List, Optional
from absl import logging

from inductiva import types
from inductiva.simulation import Simulator
from inductiva.utils import templates
from inductiva.utils import files


def explore_design_space(simulator: Simulator,
                         input_dir: types.Path,
                         template_filename: str,
                         tag: str,
                         values: List[Any],
                         output_dir: Optional[types.Path] = None,
                         **extra_sim_kwargs):
    """Explore over design parameters with a template file.

    Args:
        simulator: Simulator to use.
        input_dir: Directory where the input files (including the template
            file) are located.
        template_filename: Name of the template file.
        tag: Name of the parameter to explore.
        values: Values of the parameter to explore.
        output_dir: Directory where the output files will be stored.
        extra_sim_kwargs: Extra keyword arguments to pass to the simulator.
    """
    input_dir = files.resolve_path(input_dir)
    file_format = os.path.splitext(template_filename)[1]

    if output_dir is None:
        output_dir = input_dir.with_name(f"{input_dir.name}-output")
    else:
        output_dir = files.resolve_path(output_dir)

    logging.info("Exploring design space for attribute \"%s\".", tag)

    for value in values:
        logging.info("Running simulation for %s=%s.", tag, value)

        input_filename = f"input_file_{value}{file_format}"
        input_file = os.path.join(input_dir, input_filename)

        templates.replace_params_in_template(
            str(input_dir),
            template_filename,
            {tag: value},
            input_file,
        )

        simulator.run(
            input_dir=input_dir,
            sim_config_filename=input_filename,
            output_dir=output_dir.joinpath(f"design_{tag}={value}"),
            **extra_sim_kwargs,
        )



"""Linear algebra methods that are computed using CuPy via the API."""
import numpy as np
import scipy
import scipy.sparse
from inductiva.api import invoke_api_from_fn_ptr

from typing import Tuple
# pylint: disable=unused-argument


def eigs(m: scipy.sparse) -> Tuple[np.ndarray, np.ndarray]:
    return invoke_api_from_fn_ptr(locals(), eigs)


# pylint: enable=unused-argument



"""GROMACS module of the API"""

import pathlib
from typing import Optional

from inductiva import types
from inductiva.simulation import Simulator


class GROMACS(Simulator):
    """Class to invoke any GROMACS command on the API."""

    @property
    def api_method_name(self) -> str:
        return "md.gromacs.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        method_name: str,
        track_logs: bool = False,
        output_directory: Optional[types.Path] = None,
        **gromacs_flags: Optional[str],
    ) -> pathlib.Path:
        """Run a specified gromacs method.

        Args:
            input_dir: Path to the directory containing the input files.
            method_name: Method to run.
            output_dir: Path to the directory where the output files will be
                stored.
            gromacs_flags: Flags to pass to the gromacs CLI.
        """
        return super().run(input_dir,
                           output_dir=output_directory,
                           track_logs=track_logs,
                           method=method_name,
                           user_flags=gromacs_flags)

    def run_async(
        self,
        input_dir: types.Path,
        method_name: str,
        **gromacs_flags: Optional[str],
    ) -> str:
        """Run a specified gromacs method asynchronously.

        Args:
            input_dir: Path to the directory containing the input files.
            method_name: Method to run.
            gromacs_flags: Flags to pass to the gromacs CLI.
        """
        return super().run_async(input_dir,
                                 method=method_name,
                                 user_flags=gromacs_flags)



import typing_extensions

from inductiva.client.apis.tags import TagValues
from inductiva.client.apis.tags.admin_api import AdminApi
from inductiva.client.apis.tags.tasks_api import TasksApi

TagToApi = typing_extensions.TypedDict('TagToApi', {
    TagValues.ADMIN: AdminApi,
    TagValues.TASKS: TasksApi,
})

tag_to_api = TagToApi({
    TagValues.ADMIN: AdminApi,
    TagValues.TASKS: TasksApi,
})



# coding: utf-8

# flake8: noqa

# import all models into this package
# if you have many models here with many references from one model to another this may
# raise a RecursionError
# to avoid this, import only the models that you directly need like:
# from inductiva.client.model.pet import Pet
# or import this package, but before doing it, use:
# import sys
# sys.setrecursionlimit(n)

from inductiva.client.model.body_upload_task_input import BodyUploadTaskInput
from inductiva.client.model.http_validation_error import HTTPValidationError
from inductiva.client.model.new_user import NewUser
from inductiva.client.model.queue_status import QueueStatus
from inductiva.client.model.task_request import TaskRequest
from inductiva.client.model.task_status import TaskStatus
from inductiva.client.model.validation_error import ValidationError



"""Functions that perform dummy operations on the API useful for testing."""
from inductiva.api import invoke_api_from_fn_ptr
# pylint: disable=unused-argument, disable=redefined-builtin


def sleep(secs: float) -> None:
    return invoke_api_from_fn_ptr(locals(), sleep)


# pylint: enable=unused-argument, enable=redefined-builtin



"""Functionality to interact with the Web API."""
from .methods import invoke_api, invoke_api_from_fn_ptr, run_simulation, run_async_simulation
from .methods import download_output, invoke_async_api, get_task_status
from .methods import validate_api_key



#pylint: disable=missing-module-docstring
from .fluid_types import (
    FluidType,
    WATER,
    OLIVE_OIL,
    LIQUID_PROPANE,
    JET_FUEL,
    GEAR_OIL,
    BEER,
    HONEY,
)
from .scenarios import DamBreak
from .scenarios import FluidBlock
from .scenarios import WindTunnel
from .scenarios import (
    FluidTank,
    CubicTankOutlet,
    CylindricalTankOutlet,
    RectangularTankInlet,
    CircularTankInlet,
)
from . import scenarios
from .scenarios._post_processing import SPHSimulationOutput
from .simulators import (
    SWASH,
    XBeach,
    DualSPHysics,
    SPlisHSPlasH,
    OpenFOAM,
)



import typing_extensions

from inductiva.client.paths import PathValues
from inductiva.client.apis.paths.task_submit import TaskSubmit
from inductiva.client.apis.paths.task_task_id_input import TaskTaskIdInput
from inductiva.client.apis.paths.task_task_id_status import TaskTaskIdStatus
from inductiva.client.apis.paths.task_task_id_output import TaskTaskIdOutput
from inductiva.client.apis.paths.task_task_id_kill import TaskTaskIdKill
from inductiva.client.apis.paths.admin_user import AdminUser
from inductiva.client.apis.paths.admin_user_username_tasks import AdminUserUsernameTasks

PathToApi = typing_extensions.TypedDict(
    'PathToApi', {
        PathValues.TASK_SUBMIT: TaskSubmit,
        PathValues.TASK_TASK_ID_INPUT: TaskTaskIdInput,
        PathValues.TASK_TASK_ID_STATUS: TaskTaskIdStatus,
        PathValues.TASK_TASK_ID_OUTPUT: TaskTaskIdOutput,
        PathValues.TASK_TASK_ID_KILL: TaskTaskIdKill,
        PathValues.ADMIN_USER: AdminUser,
        PathValues.ADMIN_USER_USERNAME_TASKS: AdminUserUsernameTasks,
    })

path_to_api = PathToApi({
    PathValues.TASK_SUBMIT: TaskSubmit,
    PathValues.TASK_TASK_ID_INPUT: TaskTaskIdInput,
    PathValues.TASK_TASK_ID_STATUS: TaskTaskIdStatus,
    PathValues.TASK_TASK_ID_OUTPUT: TaskTaskIdOutput,
    PathValues.TASK_TASK_ID_KILL: TaskTaskIdKill,
    PathValues.ADMIN_USER: AdminUser,
    PathValues.ADMIN_USER_USERNAME_TASKS: AdminUserUsernameTasks,
})



# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from inductiva.client.apis.path_to_api import path_to_api

import enum


class PathValues(str, enum.Enum):
    TASK_SUBMIT = "/task/submit"
    TASK_TASK_ID_INPUT = "/task/{task_id}/input"
    TASK_TASK_ID_STATUS = "/task/{task_id}/status"
    TASK_TASK_ID_OUTPUT = "/task/{task_id}/output"
    TASK_TASK_ID_KILL = "/task/{task_id}/kill"
    ADMIN_USER = "/admin/user"
    ADMIN_USER_USERNAME_TASKS = "/admin/user/{username}/tasks"
